from pathlib import Path

from datetime import datetime, timedelta
from typing import List

import pytest
from cgmodels.cg.constants import Pipeline

from cg.constants.constants import FileFormat
from cg.io.controller import ReadFile
from cg.meta.report.balsamic import BalsamicReportAPI
from cg.meta.report.mip_dna import MipDNAReportAPI
from cg.models.cg_config import CGConfig
from cg.store.models import Family
from tests.apps.scout.conftest import MockScoutApi
from tests.mocks.balsamic_analysis_mock import MockBalsamicAnalysis
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.mip_analysis_mock import MockMipAnalysis
from tests.mocks.report import MockChanjo, MockDB, MockHousekeeperMipDNAReportAPI


@pytest.fixture(scope="function", name="report_api_mip_dna")
def report_api_mip_dna(cg_context: CGConfig, lims_samples) -> MipDNAReportAPI:
    """MIP DNA ReportAPI fixture."""

    cg_context.meta_apis["analysis_api"] = MockMipAnalysis()
    cg_context.status_db_ = MockDB(report_store)
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.chanjo_api_ = MockChanjo()
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return MockHousekeeperMipDNAReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function", name="report_api_balsamic")
def report_api_balsamic(cg_context: CGConfig, lims_samples) -> BalsamicReportAPI:
    """BALSAMIC ReportAPI fixture."""

    cg_context.meta_apis["analysis_api"] = MockBalsamicAnalysis(cg_context)
    cg_context.status_db_ = MockDB(report_store)
    cg_context.lims_api_ = MockLimsAPI(cg_context, lims_samples)
    cg_context.scout_api_ = MockScoutApi(cg_context)
    return BalsamicReportAPI(cg_context, cg_context.meta_apis["analysis_api"])


@pytest.fixture(scope="function", name="case_mip_dna")
def case_mip_dna(case_id, report_api_mip_dna) -> Family:
    """MIP DNA case instance."""

    return report_api_mip_dna.status_db.family(case_id)


@pytest.fixture(scope="function", name="case_balsamic")
def case_balsamic(case_id, report_api_balsamic) -> Family:
    """BALSAMIC case instance."""

    return report_api_balsamic.status_db.family(case_id)


@pytest.fixture(scope="function", name="case_samples_data")
def case_samples_data(case_id, report_api_mip_dna):
    """MIP DNA family sample object."""

    return report_api_mip_dna.status_db.family_samples(case_id)


@pytest.fixture(name="mip_analysis_api")
def mip_analysis_api() -> MockMipAnalysis:
    """MIP analysis mock data."""

    return MockMipAnalysis()


@pytest.fixture(name="lims_family")
def fixture_lims_family(fixtures_dir: Path) -> dict:
    """Returns a lims-like case of samples."""
    return ReadFile.get_content_from_file(
        file_format=FileFormat.JSON, file_path=Path(fixtures_dir, "report", "lims_family.json")
    )


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a lims case."""

    return lims_family["samples"]


@pytest.fixture(scope="function", autouse=True, name="report_store")
def report_store(analysis_store, helpers, timestamp_yesterday):
    """A mock store instance for report testing."""

    case = analysis_store.families()[0]
    helpers.add_analysis(
        analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=timestamp_yesterday
    )
    helpers.add_analysis(analysis_store, case, pipeline=Pipeline.MIP_DNA, started_at=datetime.now())

    # Mock sample dates to calculate processing times
    for family_sample in analysis_store.family_samples(case.internal_id):
        family_sample.sample.ordered_at = timestamp_yesterday - timedelta(days=2)
        family_sample.sample.received_at = timestamp_yesterday - timedelta(days=1)
        family_sample.sample.prepared_at = timestamp_yesterday
        family_sample.sample.sequenced_at = timestamp_yesterday
        family_sample.sample.delivered_at = datetime.now()

    return analysis_store



import pytest
from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.exc import OrderError
from cg.meta.orders.sars_cov_2_submitter import SarsCov2Submitter
from cg.models.orders.constants import OrderType
from cg.models.orders.order import OrderIn
from cg.models.orders.sample_base import ControlEnum
from cg.models.orders.samples import SarsCov2Sample
from cg.store import Store
from cg.store.models import Sample
from cg.store.models import Customer
from tests.store_helpers import StoreHelpers
import datetime as dt


def test_store_items_in_status_control_has_stored_value(
    sarscov2_order_to_submit: dict, base_store: Store
):
    # GIVEN sarscov2 order with three samples with control value
    order: OrderIn = OrderIn.parse_obj(sarscov2_order_to_submit, OrderType.SARS_COV_2)
    control_value = ControlEnum.positive
    sample: SarsCov2Sample
    for sample in order.samples:
        sample.control: ControlEnum = control_value
    submitter: SarsCov2Submitter = SarsCov2Submitter(status=base_store, lims=None)
    status_data = submitter.order_to_status(order=order)

    # WHEN storing the order
    submitter.store_items_in_status(
        comment="",
        customer_id=order.customer,
        data_analysis=Pipeline.SARS_COV_2,
        data_delivery=DataDelivery.FASTQ,
        order="",
        ordered=dt.datetime.now(),
        ticket_id=123456,
        items=status_data.get("samples"),
    )

    # THEN control should exist on the sample in the store
    customer: Customer = base_store.get_customer_by_customer_id(customer_id=order.customer)
    sample: SarsCov2Sample
    for sample in order.samples:
        stored_sample: Sample = base_store.find_samples(customer=customer, name=sample.name).first()
        assert stored_sample.control == control_value



from typing import Optional

from sqlalchemy import Column, ForeignKey, orm, types
from sqlalchemy.orm.exc import NoResultFound

from .base import Model


class Sample(Model):
    sample_id = Column(types.Integer, primary_key=True)
    project_id = Column(ForeignKey("project.project_id", ondelete="CASCADE"), nullable=False)
    samplename = Column(types.String(255), nullable=False)
    customerid = Column(types.String(255))
    limsid = Column(types.String(255))
    barcode = Column(types.String(255))
    time = Column(types.DateTime)

    unaligned = orm.relationship(
        "Unaligned", cascade="all, delete-orphan", backref=orm.backref("sample")
    )

    @property
    def lims_id(self) -> str:
        """Parse out the LIMS id from the sample name in demux database."""
        sample_part = self.samplename.split("_")[0]
        return sample_part.rstrip("FB")

    @staticmethod
    def exists(sample_name: str, barcode: str) -> Optional[int]:
        """Checks if a Sample entry already exists"""
        try:
            sample: Sample = (
                Sample.query.filter_by(samplename=sample_name).filter_by(barcode=barcode).one()
            )
            return sample.sample_id
        except NoResultFound:
            return None



"""Module for Tower Analysis API."""

import logging
from typing import List

from cg.utils.utils import build_command_from_dict

LOG = logging.getLogger(__name__)


class TowerAnalysisAPI:
    """Handles communication between tower processes
    and the rest of CG infrastructure."""

    @classmethod
    def get_tower_launch_parameters(cls, tower_pipeline: str, command_args: dict) -> List[str]:
        """Returns a tower launch command given a dictionary with arguments."""

        tower_options: List[str] = build_command_from_dict(
            options=dict(
                (f"--{arg}", command_args.get(arg, None))
                for arg in (
                    "work-dir",
                    "profile",
                    "params-file",
                    "config",
                    "name",
                    "revision",
                    "compute-env",
                )
            ),
            exclude_true=True,
        )
        parameters = ["launch"] + tower_options + [tower_pipeline]
        return parameters



import click

from cg.constants.rnafusion import RnafusionDefaults

OPTION_FROM_START = click.option(
    "--from_start",
    is_flag=True,
    default=False,
    show_default=True,
    help="Start pipeline from start without resuming execution",
)

OPTION_STRANDEDNESS = click.option(
    "--strandedness",
    type=str,
    default=RnafusionDefaults.STRANDEDNESS,
    show_default=True,
    help="Strandedness: forward, unstranded or reverse (default)",
)

OPTION_REFERENCES = click.option(
    "--genomes_base", type=click.Path(), help="Path to references folder"
)
OPTION_TRIM = click.option(
    "--trim",
    is_flag=True,
    default=RnafusionDefaults.TRIM,
    show_default=True,
    help="Preform trimming of reads to 75 bp",
)
OPTION_FUSIONINSPECTOR_FILTER = click.option(
    "--fusioninspector_filter",
    is_flag=True,
    default=RnafusionDefaults.FUSIONINSPECTOR_FILTER,
    show_default=True,
    help="Feed filtered fusionreport fusions to fusioninspector",
)
OPTION_ALL = click.option(
    "--all",
    is_flag=True,
    default=RnafusionDefaults.ALL,
    show_default=True,
    help="Run all analysis tools",
)
OPTION_PIZZLY = click.option(
    "--pizzly",
    is_flag=True,
    default=RnafusionDefaults.PIZZLY,
    show_default=True,
    help="Run pizzly analysis tool",
)
OPTION_SQUID = click.option(
    "--squid",
    is_flag=True,
    default=RnafusionDefaults.SQUID,
    show_default=True,
    help="Run squid analysis tool",
)
OPTION_STARFUSION = click.option(
    "--starfusion",
    is_flag=True,
    default=RnafusionDefaults.STARFUSION,
    show_default=True,
    help="Run starfusion analysis tool",
)
OPTION_FUSIONCATCHER = click.option(
    "--fusioncatcher",
    is_flag=True,
    default=RnafusionDefaults.FUSIONCATCHER,
    show_default=True,
    help="Run fusioncatcher analysis tool",
)
OPTION_ARRIBA = click.option(
    "--arriba",
    is_flag=True,
    default=RnafusionDefaults.ARRIBA,
    show_default=True,
    help="Run arriba analysis tool",
)



AUTHORS = [
    "Jan Albert",
    "Tobias Allander",
    "Sandra Broddesson",
    "Robert Dyrdak",
    "Martin Ekman",
    "Lynda Eneh",
    "Shambhu Ganeshappa Aralaguppe",
    "Natalija Gerasimcik",
    "Karina Hentrich",
    "Tanja Normark",
    "Henning Onsbring",
    "Maria Ropat",
    "Annika Tiveljung Lindell",
    "Valtteri Wirta",
    "Zhibing Yun",
]

HEADERS = [
    "submitter",
    "fn",
    "covv_virus_name",
    "covv_type",
    "covv_passage",
    "covv_collection_date",
    "covv_location",
    "covv_host",
    "covv_gender",
    "covv_patient_age",
    "covv_patient_status",
    "covv_seq_technology",
    "covv_orig_lab",
    "covv_orig_lab_addr",
    "covv_subm_sample_id",
    "covv_subm_lab",
    "covv_subm_lab_addr",
    "covv_authors",
    "covv_specimen",
    "covv_outbreak",
    "covv_add_host_info",
    "covv_add_location",
    "covv_provider_sample_id",
    "covv_last_vaccinated",
    "covv_treatment",
    "covv_assembly_method",
    "covv_coverage",
]



from enum import Enum
from typing import List, Optional, Callable

from sqlalchemy.orm import Query

from cg.store.models import Customer


def filter_customer_by_customer_internal_id(customers: Query, customer_id: str, **kwargs) -> Query:
    """Return customer by customer internal id."""
    return customers.filter(Customer.internal_id == customer_id)


class CustomerFilter(Enum):
    """Define customer filter functions."""

    FILTER_BY_INTERNAL_ID: Callable = filter_customer_by_customer_internal_id


def apply_customer_filter(
    customers: Query,
    filter_functions: List[Callable],
    customer_id: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for filter_function in filter_functions:
        customers: Query = filter_function(
            customers=customers,
            customer_id=customer_id,
        )
    return customers



import pytest

from datetime import datetime
from pathlib import Path
from typing import List

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.meta.demultiplex.delete_demultiplex_api import DeleteDemuxAPI
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flow_cell import FlowCell
from cg.store.api import Store
from cg.store.models import Sample, Family

from tests.apps.cgstats.conftest import fixture_populated_stats_api
from tests.cli.demultiplex.conftest import (
    fixture_demultiplex_configs,
    fixture_demultiplex_context,
    fixture_demultiplexed_flow_cell_working_directory,
    fixture_demultiplexed_flow_cells_working_directory,
    fixture_demultiplexing_api,
    fixture_demux_results_not_finished_dir,
    fixture_flow_cell_runs_working_directory,
    fixture_stats_api,
)
from tests.models.demultiplexing.conftest import (
    fixture_flowcell_path,
    fixture_flow_cell_runs,
)
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="tmp_demulitplexing_dir")
def fixture_tmp_demulitplexing_dir(
    demultiplexed_flow_cells_working_directory: Path, flow_cell_full_name: str
) -> Path:
    """Return a tmp directory in demultiplexed-runs."""
    tmp_demulitplexing_dir: Path = Path(
        demultiplexed_flow_cells_working_directory, flow_cell_full_name
    )
    tmp_demulitplexing_dir.mkdir(exist_ok=True, parents=True)
    return tmp_demulitplexing_dir


@pytest.fixture(name="tmp_fastq_paths")
def fixture_temp_fastq_paths(tmp_demulitplexing_dir: Path) -> List[Path]:
    """Return a list of temporary dummy fastq paths."""
    fastqs = [
        Path(tmp_demulitplexing_dir, "fastq_1.fastq.gz"),
        Path(tmp_demulitplexing_dir, "fastq_2.fastq.gz"),
    ]
    for fastq in fastqs:
        with fastq.open("w+") as fh:
            fh.write("content")
    return fastqs


@pytest.fixture(name="tmp_sample_sheet_path")
def fixture_tmp_samplesheet_path(tmp_demulitplexing_dir: Path) -> Path:
    """Return SampleSheet in temporary demuliplexing folder."""
    tmp_sample_sheet_path = Path(tmp_demulitplexing_dir, "SampleSheet.csv")
    with tmp_sample_sheet_path.open("w+") as fh:
        fh.write("content")
    return tmp_sample_sheet_path


@pytest.fixture(name="tmp_flow_cell_run_path")
def fixture_tmp_flow_cell_run_path(project_dir: Path, flow_cell_full_name: str) -> Path:
    """Flow cell run directory in temporary folder."""

    tmp_flow_cell_run_path: Path = Path(project_dir, "flow_cell_run", flow_cell_full_name)
    tmp_flow_cell_run_path.mkdir(exist_ok=True, parents=True)

    return tmp_flow_cell_run_path


@pytest.fixture(name="cgstats_select_project_log_file")
def fixture_cgstats_select_project_log_file(flow_cell: FlowCell, flow_cell_project_id: int) -> Path:
    """Return cgstats select project out file."""
    return Path(
        flow_cell.path,
        "-".join(["stats", str(flow_cell_project_id), flow_cell.id]) + ".txt",
    )


@pytest.fixture(name="flow_cell_project_id")
def fixture_flow_cell_project_id() -> int:
    """Return flow cell run project id."""
    return 174578


@pytest.fixture(name="hiseq_x_copy_complete_file")
def fixture_hiseq_x_copy_complete_file(flow_cell: FlowCell) -> Path:
    """Return Hiseq X flow cell copy complete file."""
    return Path(flow_cell.path, DemultiplexingDirsAndFiles.Hiseq_X_COPY_COMPLETE)


@pytest.fixture(name="populated_flow_cell_store")
def fixture_populated_flow_cell_store(
    family_name: str, flow_cell_id: str, sample_id: str, store: Store, helpers: StoreHelpers
) -> Store:
    """Populate a store with a NovaSeq flow cell."""

    populated_flow_cell_store: Store = store
    sample: Sample = helpers.add_sample(store=populated_flow_cell_store, internal_id=sample_id)
    family: Family = helpers.add_case(store=populated_flow_cell_store, internal_id=family_name)
    helpers.add_relationship(
        store=populated_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flowcell(
        store=populated_flow_cell_store,
        flow_cell_id=flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return populated_flow_cell_store


@pytest.fixture(name="active_flow_cell_store")
def fixture_active_flow_cell_store(
    family_name: str, flow_cell_id: str, sample_id: str, base_store: Store, helpers: StoreHelpers
) -> Store:
    """Populate a store with a Novaseq flow cell, with active samples on it."""
    active_flow_cell_store: Store = base_store
    sample: Sample = helpers.add_sample(store=active_flow_cell_store, internal_id=sample_id)
    family: Family = helpers.add_case(
        store=active_flow_cell_store, internal_id=family_name, action="running"
    )
    helpers.add_relationship(
        store=active_flow_cell_store,
        sample=sample,
        case=family,
    )
    helpers.add_flowcell(
        store=active_flow_cell_store,
        flow_cell_id=flow_cell_id,
        sequencer_type="novaseq",
        samples=[sample],
    )
    return active_flow_cell_store


@pytest.fixture(name="sample_level_housekeeper_api")
def fixture_sample_level_housekeeper_api(
    flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: List[Path],
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    sample_level_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    helpers.ensure_hk_bundle(store=sample_level_housekeeper_api, bundle_data=bundle_data)
    return sample_level_housekeeper_api


@pytest.fixture(name="flow_cell_name_housekeeper_api")
def fixture_flow_cell_name_housekeeper_api(
    flow_cell_id: str,
    real_housekeeper_api: HousekeeperAPI,
    sample_id: str,
    tmp_fastq_paths: List[Path],
    tmp_sample_sheet_path: Path,
    helpers,
) -> HousekeeperAPI:
    """Return a mocked Housekeeper API, containing a sample bundle with related FASTQ files."""
    flow_cell_housekeeper_api = real_housekeeper_api
    bundle_data = {
        "name": sample_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {"path": path.as_posix(), "tags": ["fastq", flow_cell_id], "archive": False}
            for path in tmp_fastq_paths
        ],
    }
    flow_cell_bundle_data = {
        "name": flow_cell_id,
        "created": datetime.now(),
        "version": "1.0",
        "files": [
            {
                "path": tmp_sample_sheet_path.as_posix(),
                "tags": ["samplesheet", flow_cell_id],
                "archive": False,
            }
        ],
    }

    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=bundle_data)
    helpers.ensure_hk_bundle(store=flow_cell_housekeeper_api, bundle_data=flow_cell_bundle_data)
    return flow_cell_housekeeper_api


@pytest.fixture(name="populated_wipe_demux_context")
def fixture_populated_wipe_demux_context(
    cg_context: CGConfig,
    flow_cell_name_housekeeper_api: HousekeeperAPI,
    populated_flow_cell_store: Store,
    populated_stats_api: StatsAPI,
) -> CGConfig:
    """Return a populated context to remove flow cells from using the DeleteDemuxAPI."""
    populated_wipe_demux_context = cg_context
    populated_wipe_demux_context.cg_stats_api_ = populated_stats_api
    populated_wipe_demux_context.status_db_ = populated_flow_cell_store
    populated_wipe_demux_context.housekeeper_api_ = flow_cell_name_housekeeper_api
    return populated_wipe_demux_context


@pytest.fixture(name="active_wipe_demux_context")
def fixture_active_wipe_demux_context(
    cg_context: CGConfig, active_flow_cell_store: Store
) -> CGConfig:
    """Return a populated context to remove flow cells from using the DeleteDemuxAPI."""
    active_wipe_demux_context = cg_context
    active_wipe_demux_context.status_db_ = active_flow_cell_store
    return active_wipe_demux_context


@pytest.fixture(name="populated_wipe_demultiplex_api")
def fixture_populated_wipe_demultiplex_api(
    populated_wipe_demux_context: CGConfig,
    demultiplexed_flow_cells_working_directory: Path,
    tmp_flow_cell_run_path: Path,
) -> DeleteDemuxAPI:
    """Return an initialized populated DeleteDemuxAPI."""
    return DeleteDemuxAPI(
        config=populated_wipe_demux_context,
        demultiplex_base=demultiplexed_flow_cells_working_directory,
        dry_run=False,
        run_path=tmp_flow_cell_run_path,
    )


@pytest.fixture(name="active_wipe_demultiplex_api")
def fixture_active_wipe_demultiplex_api(
    active_wipe_demux_context: CGConfig,
    demultiplexed_flow_cells_working_directory: Path,
    flow_cell_full_name: str,
) -> DeleteDemuxAPI:
    """Return an instantiated DeleteDemuxAPI with active samples on a flow cell."""
    return DeleteDemuxAPI(
        config=active_wipe_demux_context,
        demultiplex_base=demultiplexed_flow_cells_working_directory,
        dry_run=False,
        run_path=Path(flow_cell_full_name),
    )


@pytest.fixture(name="wipe_demultiplex_api")
def fixture_wipe_demultiplex_api(
    cg_context: CGConfig,
    demultiplexed_flow_cells_working_directory: Path,
    flow_cell_full_name: str,
    stats_api: StatsAPI,
) -> DeleteDemuxAPI:
    """Return an initialized DeleteDemuxAPI."""
    cg_context.cg_stats_api_ = stats_api
    return DeleteDemuxAPI(
        config=cg_context,
        demultiplex_base=demultiplexed_flow_cells_working_directory,
        dry_run=False,
        run_path=Path(flow_cell_full_name),
    )


@pytest.fixture(name="demultiplexing_init_files")
def tmp_demultiplexing_init_files(
    flow_cell_id: str, populated_wipe_demultiplex_api: DeleteDemuxAPI
) -> List[Path]:
    """Return a list of demultiplexing init files present in the run directory."""
    run_path: Path = populated_wipe_demultiplex_api.run_path
    slurm_job_id_file_path: Path = Path(run_path, "slurm_job_ids.yaml")
    demux_script_file_path: Path = Path(run_path, "demux-novaseq.sh")
    error_log_path: Path = Path(run_path, f"{flow_cell_id}_demultiplex.stderr")
    log_path: Path = Path(run_path, f"{flow_cell_id}_demultiplex.stdout")

    demultiplexing_init_files: List[Path] = [
        slurm_job_id_file_path,
        demux_script_file_path,
        error_log_path,
        log_path,
    ]

    for file in demultiplexing_init_files:
        file.touch()
    return demultiplexing_init_files



import logging

from housekeeper.store.models import Version

from cg.apps.lims import LimsAPI
from cg.constants.constants import PrepCategory
from cg.constants.scout_upload import RNAFUSION_CASE_TAGS, RNAFUSION_SAMPLE_TAGS, GenomeBuild
from cg.meta.upload.scout.hk_tags import CaseTags, SampleTags
from cg.meta.upload.scout.scout_config_builder import ScoutConfigBuilder
from cg.models.scout.scout_load_config import RnafusionLoadConfig, ScoutCancerIndividual
from cg.store.models import Analysis, FamilySample

LOG = logging.getLogger(__name__)


class RnafusionConfigBuilder(ScoutConfigBuilder):
    """Class for handling rnafusion information and files to be included in Scout upload."""

    def __init__(self, hk_version_obj: Version, analysis_obj: Analysis, lims_api: LimsAPI):
        super().__init__(
            hk_version_obj=hk_version_obj, analysis_obj=analysis_obj, lims_api=lims_api
        )
        self.case_tags: CaseTags = CaseTags(**RNAFUSION_CASE_TAGS)
        self.sample_tags: SampleTags = SampleTags(**RNAFUSION_SAMPLE_TAGS)
        self.load_config: RnafusionLoadConfig = RnafusionLoadConfig(track="cancer")

    def build_load_config(self) -> None:
        """Build a rnafusion-specific load config for uploading a case to scout."""
        LOG.info("Build load config for rnafusion case")
        self.add_common_info_to_load_config()
        self.load_config.human_genome_build = GenomeBuild.hg38
        self.include_case_files()

        LOG.info("Building samples")
        db_sample: FamilySample

        for db_sample in self.analysis_obj.family.links:
            self.load_config.samples.append(self.build_config_sample(case_sample=db_sample))

    def include_case_files(self) -> None:
        """Include case level files for rnafusion case."""
        LOG.info("Including RNAFUSION specific case level files")
        for scout_key in RNAFUSION_CASE_TAGS.keys():
            self._include_file(scout_key)

    def _include_file(self, scout_key) -> None:
        """Include the file path associated to a scout configuration parameter if the corresponding housekeeper tags
        are found. Otherwise return None."""
        setattr(
            self.load_config,
            scout_key,
            self.fetch_file_from_hk(getattr(self.case_tags, scout_key)),
        )

    def build_config_sample(self, case_sample: FamilySample) -> ScoutCancerIndividual:
        """Build a sample with rnafusion specific information."""
        config_sample = ScoutCancerIndividual()

        self.add_common_sample_info(config_sample=config_sample, case_sample=case_sample)

        config_sample.analysis_type = PrepCategory.WHOLE_TRANSCRIPTOME_SEQUENCING.value

        return config_sample



from enum import Enum
from typing import Callable, List, Optional
from sqlalchemy.orm import Query
from cg.store.models import Panel


def filter_panel_by_abbrev(panels: Query, abbreviation: str, **kwargs) -> Query:
    """Return panel by abbreviation."""
    return panels.filter(Panel.abbrev == abbreviation)


class PanelFilter(Enum):
    """Define Panel filter functions."""

    FILTER_BY_ABBREVIATION: Callable = filter_panel_by_abbrev


def apply_panel_filter(
    panels: Query,
    filters: List[PanelFilter],
    abbreviation: Optional[str] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for filter_function in filters:
        panels: Query = filter_function(
            panels=panels,
            abbreviation=abbreviation,
        )
    return panels



from datetime import datetime
from typing import Optional, List, Callable
from enum import Enum
from cgmodels.cg.constants import Pipeline
from sqlalchemy import and_, or_
from sqlalchemy.orm import Query

from cg.constants import REPORT_SUPPORTED_DATA_DELIVERY
from cg.constants.constants import CaseActions, DataDelivery
from cg.constants.observations import (
    LOQUSDB_SUPPORTED_PIPELINES,
    LOQUSDB_MIP_SEQUENCING_METHODS,
    LOQUSDB_BALSAMIC_SEQUENCING_METHODS,
)
from cg.store.models import Analysis, Application, Customer, Family, Sample


def get_cases_has_sequence(cases: Query, **kwargs) -> Query:
    """Return cases that is not sequenced according to record in StatusDB."""
    return cases.filter(or_(Application.is_external, Sample.sequenced_at.isnot(None)))


def get_inactive_analysis_cases(cases: Query, **kwargs) -> Query:
    """Return cases which are not set or on hold."""
    return cases.filter(
        or_(
            Family.action.is_(None),
            Family.action == CaseActions.HOLD,
        )
    )


def get_new_cases(cases: Query, date: datetime, **kwargs) -> Query:
    """Return old cases compared to date."""
    cases = cases.filter(Family.created_at < date)
    return cases.order_by(Family.created_at.asc())


def get_cases_with_pipeline(cases: Query, pipeline: Pipeline = None, **kwargs) -> Query:
    """Return cases with pipeline."""
    return cases.filter(Family.data_analysis == pipeline) if pipeline else cases


def get_cases_with_loqusdb_supported_pipeline(
    cases: Query, pipeline: Pipeline = None, **kwargs
) -> Query:
    """Return Loqusdb related cases with pipeline."""
    records: Query = (
        cases.filter(Family.data_analysis == pipeline)
        if pipeline
        else cases.filter(Family.data_analysis.in_(LOQUSDB_SUPPORTED_PIPELINES))
    )

    return records.filter(Customer.loqus_upload == True)


def get_cases_with_loqusdb_supported_sequencing_method(
    cases: Query, pipeline: Pipeline = None, **kwargs
) -> Query:
    """Return cases with Loqusdb supported sequencing method."""
    supported_sequencing_methods = {
        Pipeline.MIP_DNA: LOQUSDB_MIP_SEQUENCING_METHODS,
        Pipeline.BALSAMIC: LOQUSDB_BALSAMIC_SEQUENCING_METHODS,
    }
    return (
        cases.filter(Application.prep_category.in_(supported_sequencing_methods[pipeline]))
        if pipeline
        else cases
    )


def get_cases_for_analysis(cases: Query, **kwargs) -> Query:
    """Return cases in need of analysis by:
    1. Action set to analyze or
    2. Internally created cases with no action set and no prior analysis or
    3. Cases with no action, but new sequence data
    """
    return cases.filter(
        or_(
            Family.action == CaseActions.ANALYZE,
            and_(
                Application.is_external.isnot(True),
                Family.action.is_(None),
                Analysis.created_at.is_(None),
            ),
            and_(
                Family.action.is_(None),
                Analysis.created_at < Sample.sequenced_at,
            ),
        )
    )


def get_cases_with_scout_data_delivery(cases: Query, **kwargs) -> Query:
    """Return cases containing Scout as a data delivery option."""
    return cases.filter(Family.data_delivery.contains(DataDelivery.SCOUT))


def get_report_supported_data_delivery_cases(cases: Query, **kwargs) -> Query:
    """Return cases with a valid data delivery for delivery report generation."""
    return cases.filter(Family.data_delivery.in_(REPORT_SUPPORTED_DATA_DELIVERY))


def apply_case_filter(
    cases: Query,
    filter_functions: List[Callable],
    date: Optional[datetime] = None,
    pipeline: Optional[Pipeline] = None,
) -> Query:
    """Apply filtering functions and return filtered results."""
    for function in filter_functions:
        cases: Query = function(cases=cases, date=date, pipeline=pipeline)
    return cases


class CaseFilter(Enum):
    """Define case filters."""

    GET_HAS_SEQUENCE: Callable = get_cases_has_sequence
    GET_HAS_INACTIVE_ANALYSIS: Callable = get_inactive_analysis_cases
    GET_NEW: Callable = get_new_cases
    GET_WITH_PIPELINE: Callable = get_cases_with_pipeline
    GET_WITH_LOQUSDB_SUPPORTED_PIPELINE: Callable = get_cases_with_loqusdb_supported_pipeline
    GET_WITH_LOQUSDB_SUPPORTED_SEQUENCING_METHOD: Callable = (
        get_cases_with_loqusdb_supported_sequencing_method
    )
    GET_FOR_ANALYSIS: Callable = get_cases_for_analysis
    GET_WITH_SCOUT_DELIVERY: Callable = get_cases_with_scout_data_delivery
    GET_REPORT_SUPPORTED: Callable = get_report_supported_data_delivery_cases



import datetime as dt
import pytest

from typing import Iterable, List
from cg.constants import Pipeline
from cg.constants.constants import PrepCategory
from cg.constants.priority import PriorityTerms
from cg.meta.orders.pool_submitter import PoolSubmitter
from cg.store import Store
from cg.store.api.import_func import (
    parse_application_versions,
    parse_applications,
    versions_are_same,
)
from cg.store.api.models import ApplicationSchema, ApplicationVersionSchema
from tests.store_helpers import StoreHelpers
from cg.store.models import ApplicationVersion, Pool, Sample, Invoice, Application
from tests.meta.demultiplex.conftest import fixture_populated_flow_cell_store
from cg.constants.invoice import CustomerNames


class StoreCheckers:
    @staticmethod
    def get_versions_from_store(store: Store, application_tag: str) -> List[ApplicationVersion]:
        """Gets all versions for the specified application"""

        return store.get_application_by_tag(tag=application_tag).versions

    @staticmethod
    def get_application_from_store(store: Store, application_tag: str) -> Application:
        """Gets the specified application"""

        return store.get_application_by_tag(tag=application_tag)

    @staticmethod
    def version_exists_in_store(store: Store, application: ApplicationVersionSchema):
        """Check if the given raw version exists in the store."""
        db_versions: List[Application] = StoreCheckers.get_versions_from_store(
            store=store, application_tag=application.app_tag
        )
        return any(
            versions_are_same(version_obj=db_version, application_version=application)
            for db_version in db_versions
        )

    @staticmethod
    def all_versions_exist_in_store(store: Store, excel_path: str):
        """Check if all versions in the Excel exists in the store."""
        applications: Iterable[ApplicationVersionSchema] = parse_application_versions(
            excel_path=excel_path
        )
        return all(
            StoreCheckers.version_exists_in_store(store=store, application=application)
            for application in applications
        )

    @staticmethod
    def all_applications_exists(store: Store, applications_file: str):
        """Check if all applications in the Excel exists in the store."""
        applications: Iterable[ApplicationSchema] = parse_applications(excel_path=applications_file)
        return all(
            StoreCheckers.exists_application_in_store(store=store, application_tag=application.tag)
            for application in applications
        )

    @staticmethod
    def exists_application_in_store(store: Store, application_tag: str):
        """Check if the given raw application exists in the store"""
        db_application = StoreCheckers.get_application_from_store(
            store=store, application_tag=application_tag
        )

        return db_application is not None


@pytest.fixture(name="store_checkers")
def fixture_store_checkers() -> StoreCheckers:
    return StoreCheckers()


@pytest.fixture(name="applications_store")
def fixture_applications_store(
    store: Store, application_versions_file: str, helpers: StoreHelpers
) -> Store:
    """Return a store populated with applications from excel file"""
    versions: Iterable[ApplicationVersionSchema] = parse_application_versions(
        excel_path=application_versions_file
    )

    for version in versions:
        helpers.ensure_application(store=store, tag=version.app_tag)

    return store


@pytest.fixture(name="microbial_store")
def fixture_microbial_store(store: Store, helpers: StoreHelpers) -> Store:
    """Populate a store with microbial application tags"""
    microbial_active_apptags = ["MWRNXTR003", "MWGNXTR003", "MWMNXTR003", "MWLNXTR003"]
    microbial_inactive_apptags = ["MWXNXTR003", "VWGNXTR001", "VWLNXTR001"]

    for app_tag in microbial_active_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=False)

    for app_tag in microbial_inactive_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="mic", is_archived=True)

    return store


@pytest.fixture(name="microbial_store_dummy_tag")
def fixture_microbial_store_dummy_tag(microbial_store: Store, helpers: StoreHelpers) -> Store:
    """Populate a microbial store with a extra dummy app tag"""
    helpers.ensure_application(
        store=microbial_store, tag="dummy_tag", prep_category="mic", is_archived=False
    )
    return microbial_store


@pytest.fixture(name="rml_store")
def fixture_rml_store(store: Store, helpers: StoreHelpers) -> Store:
    """Populate a store with microbial application tags"""
    active_apptags = [
        "RMLP10R300",
        "RMLP10S130",
        "RMLP15R100",
        "RMLP15R200",
        "RMLP15R400",
        "RMLP15R500",
        "RMLP15R750",
        "RMLP15R825",
        "RMLP15S100",
        "RMLP15S125",
        "RMLP15S150",
        "RMLP15S175",
        "RMLP15S200",
        "RMLP15S225",
        "RMLP15S425",
    ]
    inactive_apptags = [
        "RMLP05R800",
        "RMLP15S250",
        "RMLP15S275",
        "RMLP15S300",
        "RMLP15S325",
        "RMLP15S350",
        "RMLP15S375",
        "RMLP15S400",
        "RMLP15S450",
        "RMLP15S475",
        "RMLP15S500",
        "RMLS05R200",
        "RMLCUSR800",
        "RMLCUSS160",
    ]

    for app_tag in active_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="rml", is_archived=False)

    for app_tag in inactive_apptags:
        helpers.ensure_application(store=store, tag=app_tag, prep_category="rml", is_archived=True)

    return store


@pytest.fixture(name="rml_pool_store")
def fixture_rml_pool_store(
    case_id: str,
    customer_id: str,
    helpers,
    sample_id: str,
    store: Store,
    ticket: str,
    timestamp_now: dt.datetime,
):
    new_customer = store.add_customer(
        internal_id=customer_id,
        name="Test customer",
        scout_access=True,
        invoice_address="skolgatan 15",
        invoice_reference="abc",
    )
    store.add_commit(new_customer)

    application = store.add_application(
        tag="RMLP05R800",
        prep_category="rml",
        description="Ready-made",
        percent_kth=80,
        percent_reads_guaranteed=75,
        sequencing_depth=0,
        target_reads=800,
    )
    store.add_commit(application)

    app_version = store.add_version(
        application=application,
        version=1,
        valid_from=timestamp_now,
        prices={
            PriorityTerms.STANDARD: 12,
            PriorityTerms.PRIORITY: 222,
            PriorityTerms.EXPRESS: 123,
            PriorityTerms.RESEARCH: 12,
        },
    )
    store.add_commit(app_version)

    new_pool = store.add_pool(
        customer=new_customer,
        name="Test",
        order="Test",
        ordered=dt.datetime.now(),
        application_version=app_version,
    )
    store.add_commit(new_pool)
    new_case = helpers.add_case(
        store=store,
        internal_id=case_id,
        name=PoolSubmitter.create_case_name(ticket=ticket, pool_name="Test"),
    )
    store.add_commit(new_case)

    new_sample = helpers.add_sample(
        store=store,
        internal_id=sample_id,
        application_tag=application.tag,
        application_type=application.prep_category,
        customer_id=new_customer.id,
    )
    new_sample.application_version = app_version
    store.add_commit(new_sample)

    helpers.add_relationship(
        store=store,
        sample=new_sample,
        case=new_case,
    )

    yield store


@pytest.fixture(name="re_sequenced_sample_store")
def fixture_re_sequenced_sample_store(
    store: Store,
    another_flow_cell_id: str,
    case_id: str,
    family_name: str,
    flow_cell_id: str,
    sample_id: str,
    ticket: str,
    timestamp_now: dt.datetime,
    helpers,
) -> Store:
    """Populate a store with a Fluffy case, with a sample that has been sequenced on two flow cells."""
    re_sequenced_sample_store: Store = store
    store_case = helpers.add_case(
        store=re_sequenced_sample_store,
        internal_id=case_id,
        name=family_name,
        data_analysis=Pipeline.FLUFFY,
    )

    store_sample = helpers.add_sample(
        internal_id=sample_id,
        is_tumour=False,
        application_type=PrepCategory.READY_MADE_LIBRARY.value,
        reads=1200000000,
        store=re_sequenced_sample_store,
        original_ticket=ticket,
        sequenced_at=timestamp_now,
    )

    one_day_ahead_of_now = timestamp_now + dt.timedelta(days=1)

    helpers.add_flowcell(
        store=re_sequenced_sample_store,
        flow_cell_id=another_flow_cell_id,
        samples=[store_sample],
        date=timestamp_now,
    )

    helpers.add_flowcell(
        store=re_sequenced_sample_store,
        flow_cell_id=flow_cell_id,
        samples=[store_sample],
        date=one_day_ahead_of_now,
    )

    helpers.add_relationship(store=re_sequenced_sample_store, case=store_case, sample=store_sample)

    return re_sequenced_sample_store


@pytest.fixture(name="max_nr_of_cases")
def fixture_max_nr_of_cases() -> int:
    """Return the number of maximum number of cases"""
    return 50


@pytest.fixture(name="max_nr_of_samples")
def fixture_max_nr_of_samples() -> int:
    """Return the number of maximum number of samples"""
    return 50


@pytest.fixture(name="EXPECTED_NUMBER_OF_NOT_ARCHIVED_APPLICATIONS")
def fixture_expected_number_of_not_archived_applications() -> int:
    """Return the number of expected number of not archived applications"""
    return 4


@pytest.fixture(name="EXPECTED_NUMBER_OF_APPLICATIONS_WITH_PREP_CATEGORY")
def fixture_expected_number_of_applications_with_prep_category() -> int:
    """Return the number of expected number of applications with prep category"""
    return 7


@pytest.fixture(name="EXPECTED_NUMBER_OF_APPLICATIONS")
def fixture_expected_number_of_applications() -> int:
    """Return the number of expected number of applications with prep category"""
    return 7


@pytest.fixture(name="store_with_samples_that_have_names")
def store_with_samples_that_have_names(
    store: Store, helpers: StoreHelpers, name="sample_1"
) -> Store:
    """Return a store with two samples of which one has a name"""
    helpers.add_sample(store=store, internal_id="test_sample_1", name=name)
    helpers.add_sample(store=store, internal_id="test_sample_2")
    return store


@pytest.fixture(name="store_with_samples_subject_id_and_tumour_status")
def store_with_samples_subject_id_and_tumour_status(
    store: Store,
    helpers: StoreHelpers,
    customer_id: str = "cust123",
    subject_id: str = "test_subject",
) -> Store:
    """Return a store with two samples that have subject ids of which one is tumour"""
    helpers.add_sample(
        store=store,
        internal_id="test_sample_1",
        name="sample_1",
        subject_id=subject_id,
        is_tumour=True,
        customer_id=customer_id,
    )

    helpers.add_sample(
        store=store,
        internal_id="test_sample_2",
        name="sample_2",
        subject_id=subject_id,
        is_tumour=False,
        customer_id=customer_id,
    )
    return store


@pytest.fixture(name="pool_name_1")
def fixture_pool_name_1() -> str:
    """Return the name of the first pool."""
    return "pool_1"


@pytest.fixture(name="pool_order_1")
def fixture_pool_order_1() -> str:
    """Return the order of the first pool."""
    return "pool_order_1"


@pytest.fixture(name="store_with_multiple_pools_for_customer")
def fixture_store_with_multiple_pools_for_customer(
    store: Store,
    helpers: StoreHelpers,
    customer_id: str = CustomerNames.cust132,
) -> Store:
    """Return a store with two pools with different names for the same customer."""
    for number in range(2):
        helpers.ensure_pool(
            store=store,
            customer_id=customer_id,
            name="_".join(["pool", str(number)]),
            order="_".join(["pool", "order", str(number)]),
        )
    yield store



import logging

import click
import datetime as dt
from cgmodels.cg.constants import Pipeline

from cg.cli.workflow.commands import ARGUMENT_CASE_ID
from cg.constants.constants import DRY_RUN
from cg.store import Store
from cg.store.models import Analysis, Family
from cg.meta.workflow.analysis import AnalysisAPI

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.pass_context
def fastq(context: click.Context):
    """Function for storing fastq-cases"""
    AnalysisAPI.get_help(context)


@fastq.command("store")
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_context
def store_fastq_analysis(context: click.Context, case_id: str, dry_run: bool = False):
    """Creates an analysis object in status-db for the given fast case"""
    LOG.info("Creating an analysis for case %s", case_id)
    status_db: Store = context.obj.status_db
    case_obj: Family = status_db.family(internal_id=case_id)
    new_analysis: Analysis = status_db.add_analysis(
        pipeline=Pipeline.FASTQ,
        completed_at=dt.datetime.now(),
        primary=True,
        started_at=dt.datetime.now(),
        family_id=case_obj.id,
    )
    if dry_run:
        return
    status_db.add_commit(new_analysis)


@fastq.command("store-available")
@DRY_RUN
@click.pass_context
def store_available_fastq_analysis(context: click.Context, dry_run: bool = False):
    """Creates an analysis object in status-db for all fastq cases to be delivered"""
    status_db: Store = context.obj.status_db
    for case in status_db.cases_to_analyze(pipeline=Pipeline.FASTQ, threshold=False):
        context.invoke(store_fastq_analysis, case_id=case.internal_id, dry_run=dry_run)



from cg.store import Store
from cg.store.models import Customer
from cg.store.filters.status_customer_filters import filter_customer_by_customer_internal_id


def test_filter_customer_by_customer_id(base_store: Store, customer_id: str):
    """Test return customer by customer internal id."""
    # GIVEN a store containing customers

    # WHEN retrieving a customer
    customer: Customer = filter_customer_by_customer_internal_id(
        customers=base_store._get_query(table=Customer),
        customer_id=customer_id,
    ).first()

    # THEN a customer should be returned
    assert customer

    # THEN the internal id should match the original
    assert customer.internal_id == customer_id



"""Functions that deals with dummy samples"""
from cg.apps.lims.samplesheet import (
    LimsFlowcellSample,
    LimsFlowcellSampleBcl2Fastq,
    LimsFlowcellSampleDragen,
)


def dummy_sample_name(sample_name: str) -> str:
    """Convert a string to a dummy sample name

    Replace space and parentheses with dashes
    """
    return sample_name.replace(" ", "-").replace("(", "-").replace(")", "-")


def dummy_sample(
    flowcell: str, dummy_index: str, lane: int, name: str, bcl_converter: str
) -> LimsFlowcellSample:
    """Constructs and returns a dummy sample in novaseq sample sheet format"""
    lims_flowcell_sample = {
        "bcl2fastq": LimsFlowcellSampleBcl2Fastq,
        "dragen": LimsFlowcellSampleDragen,
    }
    return lims_flowcell_sample[bcl_converter](
        flowcell_id=flowcell,
        lane=lane,
        sample_id=dummy_sample_name(sample_name=name),
        index=dummy_index,
        sample_name="indexcheck",
        project="indexcheck",
    )



import logging
from datetime import datetime

from cg.cli.workflow.fastq.base import store_fastq_analysis, store_available_fastq_analysis
from cg.store.models import Family, Sample


def test_store_fastq_analysis(caplog, case_id: str, cli_runner, fastq_context):
    """Test for CLI command creating an analysis object for a fastq case"""
    # GIVEN a fastq context
    caplog.set_level(logging.INFO)
    case_obj: Family = fastq_context.status_db.family(internal_id=case_id)
    case_obj.analyses = []

    # WHEN the store_fastq_analysis command is invoked
    cli_runner.invoke(store_fastq_analysis, [case_id], obj=fastq_context)

    # THEN the run command should be reached
    assert len(fastq_context.status_db.analyses(family=case_obj).all()) > 0


def test_store_available_fastq_analysis(
    caplog, case_id: str, cli_runner, fastq_context, sample_id: str
):
    """Test for CLI command creating an analysis object for all fastq cases to be delivered"""
    caplog.set_level(logging.INFO)
    # GIVEN a case with no analysis, a sample that has been sequenced and a fastq context
    case_obj: Family = fastq_context.status_db.family(internal_id=case_id)
    case_obj.analyses = []
    sample_obj: Sample = fastq_context.status_db.get_sample_by_internal_id(internal_id=sample_id)
    sample_obj.sequenced_at = datetime.now()

    # WHEN the store_available_fastq_analysis command is invoked
    cli_runner.invoke(store_available_fastq_analysis, ["--dry-run"], obj=fastq_context)

    # THEN the right case should be found and the store_fastq_analysis command should be reached
    assert f"Creating an analysis for case {case_id}" in caplog.text



from pathlib import Path
from typing import Any
from requests import Response

from cg.constants.constants import FileFormat, APIMethods
from cg.io.json import read_json, write_json, write_json_stream, read_json_stream
from cg.io.yaml import read_yaml, write_yaml, read_yaml_stream, write_yaml_stream
from cg.io.api import put, post, patch, delete, get


class ReadFile:
    """Reading file using different methods."""

    read_file = {
        FileFormat.JSON: read_json,
        FileFormat.YAML: read_yaml,
    }

    @classmethod
    def get_content_from_file(cls, file_format: str, file_path: Path) -> Any:
        """Read file using file format dispatch table."""
        return cls.read_file[file_format](file_path=file_path)


class ReadStream:
    """Reading stream using different methods."""

    read_stream = {
        FileFormat.JSON: read_json_stream,
        FileFormat.YAML: read_yaml_stream,
    }

    @classmethod
    def get_content_from_stream(cls, file_format: str, stream: Any) -> Any:
        """Read stream using file format dispatch table."""
        return cls.read_stream[file_format](stream=stream)


class WriteFile:
    """Write file using different methods."""

    write_file = {
        FileFormat.JSON: write_json,
        FileFormat.YAML: write_yaml,
    }

    @classmethod
    def write_file_from_content(cls, content: Any, file_format: str, file_path: Path) -> None:
        """Write file using file format dispatch table."""
        cls.write_file[file_format](content=content, file_path=file_path)


class WriteStream:
    """Write stream using different methods."""

    write_stream = {
        FileFormat.JSON: write_json_stream,
        FileFormat.YAML: write_yaml_stream,
    }

    @classmethod
    def write_stream_from_content(cls, content: Any, file_format: str) -> str:
        """Write stream using file format dispatch table."""
        return cls.write_stream[file_format](content=content)


class APIRequest:
    """Create API requests."""

    api_request = {
        APIMethods.PUT: put,
        APIMethods.POST: post,
        APIMethods.DELETE: delete,
        APIMethods.PATCH: patch,
        APIMethods.GET: get,
    }

    @classmethod
    def api_request_from_content(
        cls, api_method: str, url: str, headers: dict, json: dict
    ) -> Response:
        return cls.api_request[api_method](url=url, headers=headers, json=json)



"""Tests for the BaseHandle class."""
from typing import Type

import pytest
from dataclasses import astuple

from alchy import Query, ModelBase

from cg.store.api.base import BaseHandler


@pytest.mark.parametrize("table", astuple(BaseHandler()))
def test__get_query(base_store, table: Type[ModelBase]):
    """Tests the _get_query function for all attributes of BaseHandler ie tables in the database."""
    assert isinstance(base_store._get_query(table=table), Query)



import logging
from pathlib import Path

import click
from cg.apps.cgstats.parsers.conversion_stats import ConversionStats
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.demux_report import create_demux_report
from cg.exc import FlowCellError
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell

LOG = logging.getLogger(__name__)


@click.command(name="report")
@click.argument("flow-cell-name")
@click.pass_obj
def create_report_cmd(context: CGConfig, flow_cell_name: str):
    """Generate a demux report and print to stdout."""
    LOG.info(f"Check demuxed flowcell {flow_cell_name}")
    demux_api: DemultiplexingAPI = context.demultiplex_api
    try:
        flow_cell: FlowCell = FlowCell(flow_cell_path=Path(demux_api.run_dir, flow_cell_name))
    except FlowCellError as error:
        raise click.Abort from error
    demux_results: DemuxResults = DemuxResults(
        demux_dir=Path(demux_api.out_dir, flow_cell_name, flow_cell=flow_cell)
    )
    conversion_stats: Path = demux_results.conversion_stats_path
    if not conversion_stats.exists():
        LOG.warning(f"Could not find conversion stats file {conversion_stats}")
        raise click.Abort
    report = create_demux_report(
        conversion_stats=ConversionStats(demux_results.conversion_stats_path)
    )
    click.echo("\n".join(report))



"""Tests the findbusinessdata part of the Cg store API related to sample model."""

from typing import List
from cg.store import Store

from cg.store.models import (
    Sample,
    Invoice,
)
from tests.store_helpers import StoreHelpers


def test_get_all_pools_and_samples_for_invoice_by_invoice_id(store: Store, helpers: StoreHelpers):
    """Test that all pools and samples for an invoice can be fetched."""

    # GIVEN a database with a pool and a sample
    pool = helpers.ensure_pool(store=store, name="pool_1")
    sample = helpers.add_sample(store=store, name="sample_1")

    # AND an invoice with the pool and sample
    invoice: Invoice = helpers.ensure_invoice(store=store, pools=[pool], samples=[sample])

    # ASSERT that there is an invoice with a pool and a sample
    assert len(invoice.pools) == 1
    assert len(invoice.samples) == 1

    # WHEN fetching all pools and samples for the invoice
    records = store.get_pools_and_samples_for_invoice_by_invoice_id(invoice_id=invoice.id)
    # THEN the pool and sample should be returned
    assert pool in records
    assert sample in records


def test_get_samples_by_subject_id(
    store_with_samples_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
    customer_id: str = "cust123",
    subject_id: str = "test_subject",
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status.get_samples()) == 2

    # ASSERT that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_customer_id(
        customer_id=customer_id
    )

    # WHEN fetching the sample by subject id and customer_id
    samples = store_with_samples_subject_id_and_tumour_status.get_samples_by_subject_id(
        subject_id=subject_id, customer_id=customer_id
    )

    # THEN two samples should be returned
    assert samples and len(samples) == 2


def test_get_samples_by_subject_id_and_is_tumour(
    store_with_samples_subject_id_and_tumour_status: Store,
    helpers: StoreHelpers,
    customer_id: str = "cust123",
    subject_id: str = "test_subject",
    is_tumour: bool = True,
):
    """Test that samples can be fetched by subject id."""
    # GIVEN a database with two samples that have a subject ID but only one is tumour

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_subject_id_and_tumour_status.get_samples()) == 2

    # ASSERT that there is a customer with the given customer id
    assert store_with_samples_subject_id_and_tumour_status.get_customer_by_customer_id(
        customer_id=customer_id
    )
    # WHEN fetching the sample by subject id and customer_id
    samples: List[
        Sample
    ] = store_with_samples_subject_id_and_tumour_status.get_samples_by_subject_id_and_is_tumour(
        subject_id=subject_id, customer_id=customer_id, is_tumour=is_tumour
    )

    # THEN two samples should be returned
    assert samples and len(samples) == 1


def test__get_sample_by_name(store_with_samples_that_have_names: Store, name="sample_1"):
    """Test that samples can be fetched by name."""
    # GIVEN a database with two samples of which one has a name

    # ASSERT that there are two samples in the store
    assert len(store_with_samples_that_have_names.get_samples()) == 2

    # WHEN fetching the sample by name
    samples: Sample = store_with_samples_that_have_names.get_sample_by_name(name=name)

    # THEN one sample should be returned
    assert samples and samples.name == name



"""
    conftest for vogue API
"""

import pytest

from cg.apps.vogue import VogueAPI

CONFIG = {"vogue": {"binary_path": "/path/to/vogue", "config_path": "vogue_config"}}


@pytest.fixture(scope="function")
def vogue_config():
    """
    vogue config fixture
    """

    _config = CONFIG

    return _config


@pytest.fixture(scope="function", name="vogue_api")
def fixture_vogue_api(process):
    """
    vogue API fixture
    """

    _vogue_api = VogueAPI(CONFIG)
    _vogue_api.process = process
    return _vogue_api


@pytest.fixture(scope="function")
def genotype_dict():
    """
    genotype document fixture
    """

    return "{}"


@pytest.fixture(scope="function", name="process")
def fixture_process():
    """ """

    return MockProcess()


class MockProcess:
    def __init__(self):
        self._stderr = ""

    def run_command(self, parameters: list):
        self._stderr = " ".join(parameters)

    def stderr_lines(self):
        return self._stderr.split("\n")



"""Test observations API methods."""

import logging
from typing import Dict

import pytest
from _pytest.logging import LogCaptureFixture
from cgmodels.cg.constants import Pipeline

from cg.apps.loqus import LoqusdbAPI
from cg.constants.observations import LoqusdbInstance, MipDNALoadParameters, LoqusdbMipCustomers
from cg.constants.sequencing import SequencingMethod
from cg.exc import LoqusdbDuplicateRecordError, LoqusdbUploadCaseError, CaseNotFoundError
from cg.meta.observations.balsamic_observations_api import BalsamicObservationsAPI
from cg.meta.observations.mip_dna_observations_api import MipDNAObservationsAPI
from cg.models.cg_config import CGConfig
from cg.models.observations.input_files import (
    MipDNAObservationsInputFiles,
    BalsamicObservationsInputFiles,
)
from cg.store import Store
from cg.store.models import Family
from cg.store.models import Customer
from tests.store_helpers import StoreHelpers


def test_observations_upload(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    nr_of_loaded_variants: int,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test upload observations method."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a mocked observations API and a list of mocked observations files
    case: Family = analysis_store.family(case_id)
    case.customer.internal_id = LoqusdbMipCustomers.KLINISK_IMMUNOLOGI.value
    mocker.patch.object(
        mip_dna_observations_api,
        "get_observations_input_files",
        return_value=observations_input_files,
    )
    mocker.patch.object(mip_dna_observations_api, "is_duplicate", return_value=False)

    # WHEN uploading the case observations to Loqusdb
    mip_dna_observations_api.upload(case)

    # THEN the case should be successfully uploaded
    assert f"Uploaded {nr_of_loaded_variants} variants to Loqusdb" in caplog.text


def test_get_loqusdb_api(
    mip_dna_observations_api: MipDNAObservationsAPI,
    loqusdb_config_dict: Dict[LoqusdbInstance, dict],
):
    """Test Loqusdb API retrieval given a Loqusdb instance."""

    # GIVEN the expected Loqusdb config dictionary

    # GIVEN a WES Loqusdb instance and an observations API
    loqusdb_instance = LoqusdbInstance.WES

    # WHEN calling the Loqusdb API get method
    loqusdb_api: LoqusdbAPI = mip_dna_observations_api.get_loqusdb_api(loqusdb_instance)

    # THEN a WES loqusdb api should be returned
    assert isinstance(loqusdb_api, LoqusdbAPI)
    assert loqusdb_api.binary_path == loqusdb_config_dict[LoqusdbInstance.WES]["binary_path"]
    assert loqusdb_api.config_path == loqusdb_config_dict[LoqusdbInstance.WES]["config_path"]


def test_is_duplicate(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    analysis_store: Store,
    mocker,
):
    """Test duplicate extraction for a case that is not in Loqusdb."""

    # GIVEN a Loqusdb instance with no case duplicates
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(mip_dna_observations_api.loqusdb_api, "get_case", return_value=None)
    mocker.patch.object(mip_dna_observations_api.loqusdb_api, "get_duplicate", return_value=False)

    # WHEN checking that a case has not been uploaded to Loqusdb
    is_duplicate: bool = mip_dna_observations_api.is_duplicate(
        case=case,
        loqusdb_api=mip_dna_observations_api.loqusdb_api,
        profile_vcf_path=observations_input_files.profile_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN there should be no duplicates in Loqusdb
    assert is_duplicate is False


def test_is_duplicate_case_output(
    case_id: str,
    observations_input_files: MipDNAObservationsInputFiles,
    mip_dna_observations_api: MipDNAObservationsAPI,
    analysis_store: Store,
):
    """Test duplicate extraction for a case that already exists in Loqusdb."""

    # GIVEN a Loqusdb instance with a duplicated case
    case: Family = analysis_store.family(case_id)

    # WHEN checking that a case has already been uploaded to Loqusdb
    is_duplicate: bool = mip_dna_observations_api.is_duplicate(
        case=case,
        loqusdb_api=mip_dna_observations_api.loqusdb_api,
        profile_vcf_path=observations_input_files.profile_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN an upload of a duplicate case should be detected
    assert is_duplicate is True


def test_is_duplicate_loqusdb_id(
    case_id: str,
    loqusdb_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    analysis_store: Store,
    mocker,
):
    """Test duplicate extraction for a case that already exists in Loqusdb."""

    # GIVEN a Loqusdb instance with a duplicated case and whose samples already have a Loqusdb ID
    case: Family = analysis_store.family(case_id)
    case.links[0].sample.loqusdb_id = loqusdb_id
    mocker.patch.object(mip_dna_observations_api.loqusdb_api, "get_case", return_value=None)
    mocker.patch.object(mip_dna_observations_api.loqusdb_api, "get_duplicate", return_value=False)

    # WHEN checking that the sample observations have already been uploaded
    is_duplicate: bool = mip_dna_observations_api.is_duplicate(
        case=case,
        loqusdb_api=mip_dna_observations_api.loqusdb_api,
        profile_vcf_path=observations_input_files.profile_vcf_path,
        profile_threshold=MipDNALoadParameters.PROFILE_THRESHOLD.value,
    )

    # THEN a duplicated upload should be identified
    assert is_duplicate is True


def test_check_customer_loqusdb_permissions(
    customer_rare_diseases: Customer,
    customer_balsamic: Customer,
    mip_dna_observations_api: MipDNAObservationsAPI,
    caplog: LogCaptureFixture,
):
    """Test customers Loqusdb permissions."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a MIP observations API, a Rare Disease customer and a Cancer customer

    # WHEN verifying the permissions for Loqusdb upload
    mip_dna_observations_api.check_customer_loqusdb_permissions(customer_rare_diseases)

    # THEN it should be only possible to upload data from a RD customer
    assert f"Valid customer {customer_rare_diseases.internal_id} for Loqusdb uploads" in caplog.text
    with pytest.raises(LoqusdbUploadCaseError):
        mip_dna_observations_api.check_customer_loqusdb_permissions(customer_balsamic)


def test_mip_dna_get_loqusdb_instance(mip_dna_observations_api: MipDNAObservationsAPI):
    """Test Loqusdb instance retrieval given a sequencing method."""

    # GIVEN a rare disease observations API with a WES as sequencing method
    mip_dna_observations_api.sequencing_method = SequencingMethod.WES

    # WHEN getting the Loqusdb instance
    loqusdb_instance = mip_dna_observations_api.get_loqusdb_instance()

    # THEN the correct loqusdb instance should be returned
    assert loqusdb_instance == LoqusdbInstance.WES


def test_mip_dna_get_loqusdb_instance_not_supported(
    mip_dna_observations_api: MipDNAObservationsAPI, caplog: LogCaptureFixture
):
    """Test Loqusdb instance retrieval given a not supported sequencing method."""

    # GIVEN a rare disease observations API with a WTS sequencing method
    mip_dna_observations_api.sequencing_method = SequencingMethod.WTS

    # WHEN getting the Loqusdb instance
    with pytest.raises(LoqusdbUploadCaseError):
        # THEN the upload should be canceled
        mip_dna_observations_api.get_loqusdb_instance()

    assert (
        f"Sequencing method {SequencingMethod.WTS} is not supported by Loqusdb. Cancelling upload."
        in caplog.text
    )


def test_mip_dna_load_observations(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    nr_of_loaded_variants: int,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test loading of case observations for rare disease."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a mock MIP DNA observations API and a list of observations input files
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(mip_dna_observations_api, "is_duplicate", return_value=False)

    # WHEN loading the case to Loqusdb
    mip_dna_observations_api.load_observations(case, observations_input_files)

    # THEN the observations should be loaded without any errors
    assert f"Uploaded {nr_of_loaded_variants} variants to Loqusdb" in caplog.text


def test_mip_dna_load_observations_duplicate(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test upload case duplicate to Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a mocked observations API and a case object that has already been uploaded to Loqusdb
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(mip_dna_observations_api, "is_duplicate", return_value=True)

    # WHEN uploading the case observations to Loqusdb
    with pytest.raises(LoqusdbDuplicateRecordError):
        # THEN a duplicate record error should be raised
        mip_dna_observations_api.load_observations(case, observations_input_files)

    assert f"Case {case.internal_id} has already been uploaded to Loqusdb" in caplog.text


def test_mip_dna_load_observations_tumor_case(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test loading of a tumor case to Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a MIP DNA observations API and a case object with a tumour sample
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(mip_dna_observations_api, "is_duplicate", return_value=False)
    case.links[0].sample.is_tumour = True

    # WHEN getting the Loqusdb API
    with pytest.raises(LoqusdbUploadCaseError):
        # THEN an upload error should be raised and the execution aborted
        mip_dna_observations_api.load_observations(case, observations_input_files)

    assert f"Case {case.internal_id} has tumour samples. Cancelling upload." in caplog.text


def test_mip_dna_delete_case(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    analysis_store: Store,
    caplog: LogCaptureFixture,
):
    """Test delete case from Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a Loqusdb instance filled with a case
    case: Family = analysis_store.family(case_id)

    # WHEN deleting a case
    mip_dna_observations_api.delete_case(case)

    # THEN the case should be deleted from Loqusdb
    assert f"Removed observations for case {case.internal_id} from Loqusdb" in caplog.text


def test_mip_dna_delete_case_not_found(
    base_context: CGConfig,
    helpers: StoreHelpers,
    loqusdb_api: LoqusdbAPI,
    mip_dna_observations_api: MipDNAObservationsAPI,
    caplog: LogCaptureFixture,
):
    """Test delete case from Loqusdb that has not been uploaded."""
    store: Store = base_context.status_db

    # GIVEN an observations instance and a case that has not been uploaded to Loqusdb
    loqusdb_api.process.stdout = None
    mip_dna_observations_api.loqusdb_api = loqusdb_api
    case: Family = helpers.add_case(store)

    # WHEN deleting a rare disease case that does not exist in Loqusdb
    with pytest.raises(CaseNotFoundError):
        # THEN a CaseNotFoundError should be raised
        mip_dna_observations_api.delete_case(case)

    assert (
        f"Case {case.internal_id} could not be found in Loqusdb. Skipping case deletion."
        in caplog.text
    )


def test_balsamic_load_observations(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    balsamic_observations_input_files: BalsamicObservationsInputFiles,
    nr_of_loaded_variants: int,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test loading of cancer case observations."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a mock BALSAMIC observations API and a list of observations input files
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(balsamic_observations_api, "is_duplicate", return_value=False)

    # WHEN loading the case to Loqusdb
    balsamic_observations_api.load_observations(case, balsamic_observations_input_files)

    # THEN the observations should be loaded successfully
    assert f"Uploaded {nr_of_loaded_variants} variants to Loqusdb" in caplog.text


def test_balsamic_load_observations_duplicate(
    case_id: str,
    mip_dna_observations_api: MipDNAObservationsAPI,
    observations_input_files: MipDNAObservationsInputFiles,
    analysis_store: Store,
    caplog: LogCaptureFixture,
    mocker,
):
    """Test upload cancer duplicate case observations to Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a balsamic observations API and a case object that has already been uploaded to Loqusdb
    case: Family = analysis_store.family(case_id)
    mocker.patch.object(mip_dna_observations_api, "is_duplicate", return_value=True)

    # WHEN uploading the case observations to Loqusdb
    with pytest.raises(LoqusdbDuplicateRecordError):
        # THEN a duplicate record error should be raised
        mip_dna_observations_api.load_observations(case, observations_input_files)

    assert f"Case {case.internal_id} has already been uploaded to Loqusdb" in caplog.text


def test_balsamic_load_cancer_observations(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    balsamic_observations_input_files: BalsamicObservationsInputFiles,
    nr_of_loaded_variants: int,
    analysis_store: Store,
    caplog: LogCaptureFixture,
):
    """Test loading of case observations for cancer."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a mock BALSAMIC observations API and a list of observations input files
    case: Family = analysis_store.family(case_id)

    # WHEN loading the case to a somatic Loqusdb instance
    balsamic_observations_api.load_cancer_observations(
        case, balsamic_observations_input_files, balsamic_observations_api.loqusdb_somatic_api
    )

    # THEN the observations should be loaded successfully
    assert f"Uploaded {nr_of_loaded_variants} variants to Loqusdb" in caplog.text
    print(caplog.text)


def test_balsamic_delete_case(
    case_id: str,
    balsamic_observations_api: BalsamicObservationsAPI,
    analysis_store: Store,
    caplog: LogCaptureFixture,
):
    """Test delete balsamic case observations from Loqusdb."""
    caplog.set_level(logging.DEBUG)

    # GIVEN a Loqusdb instance and a case that has been uploaded to both somatic and tumor instances
    case: Family = analysis_store.family(case_id)

    # WHEN deleting the case
    balsamic_observations_api.delete_case(case)

    # THEN the case should be deleted from Loqusdb
    assert f"Removed observations for case {case.internal_id} from Loqusdb" in caplog.text


def test_balsamic_delete_case_not_found(
    base_context: CGConfig,
    helpers: StoreHelpers,
    loqusdb_api: LoqusdbAPI,
    balsamic_observations_api: BalsamicObservationsAPI,
    caplog: LogCaptureFixture,
):
    """Test delete balsamic case observations from Loqusdb that have not been uploaded."""
    store: Store = base_context.status_db

    # GIVEN empty Loqusdb instances
    loqusdb_api.process.stdout = None
    balsamic_observations_api.loqusdb_somatic_api = loqusdb_api
    balsamic_observations_api.loqusdb_tumor_api = loqusdb_api
    case: Family = helpers.add_case(store)

    # WHEN deleting a cancer case that does not exist in Loqusdb
    with pytest.raises(CaseNotFoundError):
        # THEN a CaseNotFoundError should be raised
        balsamic_observations_api.delete_case(case)

    assert (
        f"Case {case.internal_id} could not be found in Loqusdb. Skipping case deletion."
        in caplog.text
    )



import logging

import alchy

from cg.store.models import Model
from cg.store.api.delete import DeleteDataHandler
from cg.store.api.find_business_data import FindBusinessDataHandler

from .add import AddHandler
from .find_basic_data import FindBasicDataHandler
from .status import StatusHandler

LOG = logging.getLogger(__name__)


class CoreHandler(
    AddHandler,
    DeleteDataHandler,
    FindBasicDataHandler,
    FindBusinessDataHandler,
    StatusHandler,
):
    """Aggregating class for the store api handlers."""

    pass


class Store(alchy.Manager, CoreHandler):
    uri: str = ""

    def __init__(self, uri):
        self.uri = uri
        super(Store, self).__init__(config=dict(SQLALCHEMY_DATABASE_URI=uri), Model=Model)



"""Fixtures for store tests."""
import datetime
import enum
from pathlib import Path
from typing import Generator
import pytest

from cg.constants import Pipeline
from cg.constants.subject import Gender
from cg.store import Store
from cg.store.models import Analysis, Family, Sample
from tests.store_helpers import StoreHelpers


class StoreConftestFixture(enum.Enum):
    INTERNAL_ID_SAMPLE_WITH_ATTRIBUTES: str = "sample_with_attributes"
    NAME_SAMPLE_WITH_ATTRIBUTES: str = "sample_with_attributes"
    SUBJECT_ID_SAMPLE_WITH_ATTRIBUTES: str = "test_subject_id"
    DOWN_SAMPLED_TO_SAMPLE_WITH_ATTRIBUTES: int = 1
    ENTRY_ID_SAMPLE_WITH_ATTRIBUTES: int = 1
    INVOICE_ID_SAMPLE_WITH_ATTRIBUTES: int = 1
    INTERNAL_ID_SAMPLE_WITHOUT_ATTRIBUTES: str = "sample_without_attributes"
    NAME_SAMPLE_WITHOUT_ATTRIBUTES: str = "sample_without_attributes"
    ENTRY_ID_SAMPLE_WITHOUT_ATTRIBUTES: int = 2

    INVOICE_ID_POOL_WITH_ATTRIBUTES: int = 1
    NAME_POOL_WITH_ATTRIBUTES: str = "pool_with_attributes"
    NAME_POOL_WITHOUT_ATTRIBUTES: str = "pool_without_attributes"
    ORDER_POOL_WITH_ATTRIBUTES: str = "order001"

    TAG_APPLICATION_WITH_ATTRIBUTES: str = "test_tag"
    PREP_CATEGORY_APPLICATION_WITH_ATTRIBUTES: str = "wgs"
    PREP_CATEGORY_APPLICATION_WITHOUT_ATTRIBUTES: str = "wes"
    TAG_APPLICATION_WITHOUT_ATTRIBUTES: str = "test_tag_2"

    INVOICE_ID_INVOICE_WITH_ATTRIBUTES: int = 1
    INVOICE_ID_INVOICE_WITHOUT_ATTRIBUTES: int = 2


@pytest.fixture(name="application_versions_file")
def fixture_application_versions_file(fixtures_dir: Path) -> str:
    """Return application version import file."""
    return Path(fixtures_dir, "store", "api", "application_versions.xlsx").as_posix()


@pytest.fixture(name="applications_file")
def fixture_applications_file(fixtures_dir: Path) -> str:
    """Return application import file."""
    return Path(fixtures_dir, "store", "api", "applications.xlsx").as_posix()


@pytest.fixture(name="microbial_submitted_order")
def fixture_microbial_submitted_order() -> dict:
    """Build an example order as it looks after submission to."""

    def _get_item(name: str, internal_id: str, well_position: str, organism: str) -> dict:
        """Return a item."""
        ref_genomes = {
            "C. Jejuni": "NC_111",
            "M. upium": "NC_222",
            "C. difficile": "NC_333",
        }
        return dict(
            name=name,
            internal_id=internal_id,
            reads="1000",
            container="96 well plate",
            container_name="hej",
            rml_plate_name=None,
            well_position=well_position,
            well_position_rml=None,
            sex=None,
            panels=None,
            require_qcok=True,
            application="MWRNXTR003",
            source=None,
            status=None,
            customer="cust015",
            family=None,
            priority="standard",
            capture_kit=None,
            comment="comment",
            index=None,
            reagent_label=None,
            tumour=False,
            custom_index=None,
            elution_buffer="Nuclease-free water",
            organism=organism,
            reference_genome=ref_genomes[organism],
            extraction_method="MagNaPure 96 (contact Clinical Genomics before " "submission)",
            analysis=str(Pipeline.FASTQ),
            concentration_sample="1",
            mother=None,
            father=None,
        )

    return {
        "customer": "cust000",
        "name": "test order",
        "internal_id": "lims_reference",
        "comment": "test comment",
        "ticket_number": "123456",
        "items": [
            _get_item("Jag", "ms1", "D:5", "C. Jejuni"),
            _get_item("testar", "ms2", "H:5", "M. upium"),
            _get_item("list", "ms3", "A:6", "C. difficile"),
        ],
        "project_type": "microbial",
    }


@pytest.fixture(name="microbial_store")
def fixture_microbial_store(
    base_store: Store, microbial_submitted_order: dict
) -> Generator[Store, None, None]:
    """Set up a microbial store instance."""
    customer: Customer = base_store.get_customer_by_customer_id(
        customer_id=microbial_submitted_order["customer"]
    )

    for sample_data in microbial_submitted_order["items"]:
        application_version = base_store.get_application_by_tag(
            sample_data["application"]
        ).versions[0]
        organism = base_store.Organism(
            internal_id=sample_data["organism"], name=sample_data["organism"]
        )
        base_store.add(organism)
        sample = base_store.add_sample(
            name=sample_data["name"],
            sex=Gender.UNKNOWN,
            comment=sample_data["comment"],
            priority=sample_data["priority"],
            reads=sample_data["reads"],
            reference_genome=sample_data["reference_genome"],
        )
        sample.application_version = application_version
        sample.customer = customer
        sample.organism = organism
        base_store.add(sample)

    base_store.commit()
    yield base_store


@pytest.fixture(name="analysis_obj")
def fixture_analysis_obj(analysis_store: Store) -> Analysis:
    """Return an analysis object from a populated store."""
    return analysis_store.analyses()[0]


@pytest.fixture(name="case_obj")
def fixture_case_obj(analysis_store: Store) -> Family:
    """Return a case models object."""
    return analysis_store.families()[0]


@pytest.fixture(name="sample_obj")
def fixture_sample_obj(analysis_store) -> Sample:
    """Return a sample models object."""
    return analysis_store.get_samples()[0]


@pytest.fixture(name="sequencer_name")
def fixture_sequencer_name() -> str:
    """Return sequencer name."""
    return "A00689"


@pytest.fixture(name="store_with_a_sample_that_has_many_attributes_and_one_without")
def fixture_store_with_a_sample_that_has_many_attributes_and_one_without(
    store: Store,
    helpers: StoreHelpers,
    timestamp_now=datetime.datetime.now(),
) -> Store:
    """Return a store with a sample that has many attributes and one without."""
    helpers.add_sample(
        store,
        internal_id=StoreConftestFixture.INTERNAL_ID_SAMPLE_WITH_ATTRIBUTES.value,
        name=StoreConftestFixture.NAME_SAMPLE_WITH_ATTRIBUTES.value,
        is_external=True,
        is_tumour=True,
        delivered_at=timestamp_now,
        received_at=timestamp_now,
        sequenced_at=timestamp_now,
        prepared_at=timestamp_now,
        subject_id=StoreConftestFixture.SUBJECT_ID_SAMPLE_WITH_ATTRIBUTES.value,
        invoice_id=StoreConftestFixture.INVOICE_ID_SAMPLE_WITH_ATTRIBUTES.value,
        downsampled_to=StoreConftestFixture.DOWN_SAMPLED_TO_SAMPLE_WITH_ATTRIBUTES.value,
        no_invoice=False,
    )

    helpers.add_sample(
        store,
        internal_id=StoreConftestFixture.INTERNAL_ID_SAMPLE_WITHOUT_ATTRIBUTES.value,
        name=StoreConftestFixture.NAME_SAMPLE_WITHOUT_ATTRIBUTES.value,
        is_external=False,
        is_tumour=False,
        delivered_at=None,
        received_at=None,
        sequenced_at=None,
        prepared_at=None,
        subject_id=None,
        invoice_id=None,
        downsampled_to=None,
        no_invoice=True,
    )

    return store


@pytest.fixture(name="store_with_a_pool_with_and_without_attributes")
def fixture_store_with_a_pool_with_and_without_attributes(
    store: Store,
    helpers: StoreHelpers,
    timestamp_now=datetime.datetime.now(),
) -> Store:
    """Return a store with a pool with and without attributes."""
    helpers.ensure_pool(
        store=store,
        delivered_at=timestamp_now,
        received_at=timestamp_now,
        invoice_id=StoreConftestFixture.INVOICE_ID_POOL_WITH_ATTRIBUTES.value,
        no_invoice=False,
        name=StoreConftestFixture.NAME_POOL_WITH_ATTRIBUTES.value,
        order=StoreConftestFixture.ORDER_POOL_WITH_ATTRIBUTES.value,
    )

    helpers.ensure_pool(
        store=store,
        delivered_at=None,
        received_at=None,
        invoice_id=None,
        no_invoice=True,
        name=StoreConftestFixture.NAME_POOL_WITHOUT_ATTRIBUTES.value,
    )

    return store


@pytest.fixture(name="store_with_an_application_with_and_without_attributes")
def fixture_store_with_an_application_with_and_without_attributes(
    store: Store,
    helpers: StoreHelpers,
    timestamp_now=datetime.datetime.now(),
) -> Store:
    """Return a store with an application with and without attributes."""
    helpers.ensure_application(
        store=store,
        tag=StoreConftestFixture.TAG_APPLICATION_WITH_ATTRIBUTES.value,
        prep_category=StoreConftestFixture.PREP_CATEGORY_APPLICATION_WITH_ATTRIBUTES.value,
        is_external=True,
        is_archived=True,
    )

    helpers.ensure_application(
        store=store,
        tag=StoreConftestFixture.TAG_APPLICATION_WITHOUT_ATTRIBUTES.value,
        prep_category=StoreConftestFixture.PREP_CATEGORY_APPLICATION_WITHOUT_ATTRIBUTES.value,
        is_external=False,
        is_archived=False,
    )

    return store


@pytest.fixture(name="store_with_an_invoice_with_and_without_attributes")
def fixture_store_with_an_invoice_with_and_without_attributes(
    store: Store,
    helpers: StoreHelpers,
    timestamp_now=datetime.datetime.now(),
) -> Store:
    """Return a store with an invoice with and without attributes."""
    helpers.ensure_invoice(
        store=store,
        invoice_id=StoreConftestFixture.INVOICE_ID_INVOICE_WITH_ATTRIBUTES.value,
        invoiced_at=timestamp_now,
    )

    helpers.ensure_invoice(
        store=store,
        invoice_id=StoreConftestFixture.INVOICE_ID_INVOICE_WITHOUT_ATTRIBUTES.value,
        invoiced_at=None,
    )

    return store



from pathlib import Path

import pytest

from cg.constants.demultiplexing import BclConverter
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flow_cell import FlowCell


@pytest.fixture(name="flow_cell_name")
def fixture_flow_cell_name(flow_cell_full_name: str) -> str:
    return flow_cell_full_name.split("_")[-1][1:]


@pytest.fixture(name="flow_cell_runs")
def fixture_flow_cell_runs(demultiplex_fixtures: Path) -> Path:
    return Path(demultiplex_fixtures, "flowcell-runs")


@pytest.fixture(name="flow_cell_path")
def fixture_flowcell_path(flow_cell_runs: Path, flow_cell_full_name: str) -> Path:
    return Path(flow_cell_runs, flow_cell_full_name)


@pytest.fixture(name="flow_cell")
def fixture_flow_cell(flow_cell_path: Path) -> FlowCell:
    flow_cell = FlowCell(flow_cell_path=flow_cell_path)
    flow_cell.parse_flow_cell_name()
    return FlowCell(flow_cell_path=flow_cell_path)


@pytest.fixture(name="dragen_flow_cell_full_name")
def fixture_dragen_flow_cell_full_name() -> str:
    return "211101_A00187_0615_AHLG5GDRXY"


@pytest.fixture(name="dragen_flow_cell_path")
def fixture_dragen_flow_cell_path(flow_cell_runs: Path, dragen_flow_cell_full_name: str) -> Path:
    return Path(flow_cell_runs, dragen_flow_cell_full_name)


@pytest.fixture(name="demultiplexed_dragen_flow_cell")
def fixture_demultiplexed_dragen_flow_cell(
    demultiplexed_runs: Path, dragen_flow_cell_full_name: str
) -> Path:
    return Path(demultiplexed_runs, dragen_flow_cell_full_name)


@pytest.fixture(name="dragen_flow_cell")
def fixture_dragen_flow_cell(dragen_flow_cell_path: Path) -> FlowCell:
    flow_cell = FlowCell(flow_cell_path=dragen_flow_cell_path)
    flow_cell.parse_flow_cell_name()
    return FlowCell(flow_cell_path=dragen_flow_cell_path)


@pytest.fixture(name="dragen_demux_results")
def fixture_dragen_demux_results(
    demultiplexed_dragen_flow_cell: Path, dragen_flow_cell: FlowCell
) -> DemuxResults:
    return DemuxResults(
        demux_dir=demultiplexed_dragen_flow_cell,
        flow_cell=dragen_flow_cell,
        bcl_converter=BclConverter.DRAGEN,
    )



from typing import List

from cg.exc import OrderError
from cg.meta.orders.microbial_submitter import MicrobialSubmitter
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import SarsCov2Sample
from cg.store.models import Customer


class SarsCov2Submitter(MicrobialSubmitter):
    """Class for validating Sars-Cov submissions."""

    def validate_order(self, order: OrderIn) -> None:
        """Validate order sample names."""
        super().validate_order(order=order)
        self._validate_sample_names_are_available(samples=order.samples, customer_id=order.customer)

    def _validate_sample_names_are_available(
        self, samples: List[SarsCov2Sample], customer_id: str
    ) -> None:
        """Validate names of all samples are not already in use."""
        customer: Customer = self.status.get_customer_by_customer_id(customer_id=customer_id)
        for sample in samples:
            if sample.control:
                continue
            if self.status.find_samples(customer=customer, name=sample.name).first():
                raise OrderError(
                    f"Sample name {sample.name} already in use for customer {customer.name}"
                )



"""
    Tests for GisaidAPI
"""

import pytest
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.exc import InvalidFastaError, HousekeeperBundleVersionMissingError
from cg.meta.upload.gisaid import GisaidAPI
from cg.meta.upload.gisaid.models import FastaFile, GisaidSample
from housekeeper.store.models import File


# def test_get_gisaid_fasta_objects_no_fasta(gisaid_api, dummy_gisaid_sample, mocker):
#    # GIVEN a hk api and a dummy gisaid sample with family_id, not represented in the hk api
#    mocker.patch.object(HousekeeperAPI, "last_version")
#    HousekeeperAPI.last_version.return_value = None
#
#    # WHEN running get_gisaid_fasta
#    # THEN HousekeeperVersionMissingError is being raised
#    with pytest.raises(HousekeeperVersionMissingError):
#        gisaid_api.get_gisaid_fasta(gisaid_samples=[dummy_gisaid_sample], case_id="dummy_case")
#
#
# def test_build_gisaid_fasta_invalid_fasta(
#    gisaid_api: GisaidAPI,
#    four_gisaid_samples: List[GisaidSample],
#    invalid_fasta_file,
#    temp_result_file: Path,
#    gisaid_case_id: str,
#    mocker,
# ):
#    # GIVEN four gisaid samples and a hk api populated with a bundle with a consensus file related to the samples
#    mocker.patch.object(HousekeeperAPI, "find_file_in_latest_version")
#    HousekeeperAPI.find_file_in_latest_version.return_value = File(
#        path=invalid_fasta_file.absolute()
#    )
#
#    # WHEN running build_gisaid_fasta with the four gisaid samples
#
#    with pytest.raises(InvalidFastaError):
#        fasta_file: Path = gisaid_api.build_gisaid_fasta(
#            gisaid_samples=four_gisaid_samples, case_id=gisaid_case_id, file_name=temp_result_file
#        )
#
#
# def test_build_gisaid_fasta(
#    gisaid_api: GisaidAPI,
#    four_gisaid_samples: List[GisaidSample],
#    valid_gisiad_fasta_file,
#    valid_housekeeper_fasta_file,
#    temp_result_file: Path,
#    gisaid_case_id: str,
#    mocker,
# ):
#    # GIVEN four gisaid samples and a hk api populated with a bundle with a consensus file related to the samples
#    mocker.patch.object(HousekeeperAPI, "find_file_in_latest_version")
#    HousekeeperAPI.find_file_in_latest_version.return_value = File(
#        path=valid_housekeeper_fasta_file.absolute()
#    )
#
#    # WHEN running build_gisaid_fasta with the four gisaid samples
#    fasta_file: Path = gisaid_api.build_gisaid_fasta(
#        gisaid_samples=four_gisaid_samples, case_id=gisaid_case_id, file_name=temp_result_file
#    )
#
#    # THEN files were found and parsed into four fasta file objects
#    assert fasta_file.read_text() == valid_gisiad_fasta_file.read_text()
#
#
# def test_get_gisaid_samples(gisaid_api: GisaidAPI, case_id: str, mocker):
#    # GIVEN a gisaid_api with a Store populated with sequenced samples with family_id=case_id
#    mocker.patch.object(LimsAPI, "get_sample_attribute")
#    LimsAPI.get_sample_attribute.side_effect = 30 * [
#        "2020-11-22",
#        "Stockholm",
#        "01",
#        "Karolinska University Hospital",
#        "171 76 Stockholm, Sweden",
#    ]
#
#    # WHEN running get_gisaid_samples
#    gisaid_samples = gisaid_api.get_gisaid_samples(case_id=case_id)
#
#    # THEN assert
#    assert gisaid_samples != []
#    for sample in gisaid_samples:
#        assert isinstance(sample, GisaidSample)
#
#
# def test_get_gisaid_samples_no_samples(gisaid_api):
#    # GIVEN a
#
#    # WHEN running get_gisaid_samples
#    gisaid_samples: GisaidAPI = gisaid_api.get_gisaid_samples(case_id="dummy_case")
#
#    # THEN assert
#    assert gisaid_samples == []
#
#
# def test_build_gisaid_csv(
#    gisaid_api, four_gisaid_samples, four_samples_csv, temp_result_file: Path
# ):
#    # GIVEN four gisaid samples
#
#    # WHEN running build_gisaid_csv
#    gisaid_api.build_gisaid_csv(gisaid_samples=four_gisaid_samples, file_name=temp_result_file)
#
#    # THEN the generated file should have the expected content
#    assert temp_result_file.read_text() == four_samples_csv
#



"""Set data in the status database and LIMS."""
import datetime
import getpass
import logging
from typing import Iterable, List, Optional

import click
from cg.cli.set.families import families
from cg.cli.set.family import family
from cg.constants import FLOWCELL_STATUS
from cg.exc import LimsDataError
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample, Customer, ApplicationVersion, Flowcell

CONFIRM = "Continue?"
HELP_KEY_VALUE = "Give a property on sample and the value to set it to, e.g. -kv name Prov52"
HELP_SKIP_LIMS = "Skip setting value in LIMS"
HELP_YES = "Answer yes on all confirmations"
OPTION_LONG_KEY_VALUE = "--key-value"
OPTION_LONG_SKIP_LIMS = "--skip-lims"
OPTION_LONG_YES = "--yes"
OPTION_SHORT_KEY_VALUE = "-kv"
OPTION_SHORT_YES = "-y"
NOT_CHANGEABLE_SAMPLE_ATTRIBUTES = [
    "application_version_id",
    "customer_id",
    "deliveries",
    "flowcells",
    "id",
    "internal_id",
    "invoice",
    "invoice_id",
    "is_external",
    "links",
    "state",
    "to_dict",
]

LOG = logging.getLogger(__name__)


@click.group("set")
def set_cmd():
    """Update information in the database."""
    pass


@set_cmd.command()
@click.option(
    "-id",
    "--identifier",
    "identifiers",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give an identifier on sample and the value to use it with, e.g. -id name Prov52",
)
@click.option(
    "-kv",
    "--key-value",
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help="Give a property on sample and the value to set it to, e.g. -kv name Prov52",
)
@click.option("--skip-lims", is_flag=True, help="Skip setting value in LIMS")
@click.option("-y", "--yes", is_flag=True, help="Answer yes on all confirmations")
@click.argument("case_id", required=False)
@click.pass_context
def samples(
    context: click.Context,
    identifiers: click.Tuple([str, str]),
    kwargs: click.Tuple([str, str]),
    skip_lims: bool,
    yes: bool,
    case_id: str,
):
    """Set values on many samples at the same time."""
    store: Store = context.obj.status_db
    sample_objs = _get_samples(case_id=case_id, identifiers=identifiers, store=store)

    if not sample_objs:
        LOG.error("No samples to alter!")
        context.abort()

    LOG.info("Would alter samples:")

    for sample_obj in sample_objs:
        LOG.info(f"{sample_obj}")

    if not (yes or click.confirm(CONFIRM)):
        raise click.Abort

    for sample_obj in sample_objs:
        context.invoke(
            sample, sample_id=sample_obj.internal_id, kwargs=kwargs, yes=yes, skip_lims=skip_lims
        )


def _get_samples(case_id: str, identifiers: click.Tuple([str, str]), store: Store) -> List[Sample]:
    """Get samples that match both case_id and identifiers if given."""
    samples_by_case_id = None
    samples_by_id = None

    if case_id:
        samples_by_case_id: List[Sample] = store.get_samples_by_case_id(case_id=case_id)

    if identifiers:
        samples_by_id = _get_samples_by_identifiers(identifiers, store)

    if case_id and identifiers:
        sample_objs = set(set(samples_by_case_id) & set(samples_by_id))
    else:
        sample_objs = samples_by_case_id or samples_by_id

    return sample_objs


def _get_samples_by_identifiers(identifiers: click.Tuple([str, str]), store: Store) -> List[Sample]:
    """Get samples matched by given set of identifiers."""
    identifier_args = {
        identifier_name: identifier_value for identifier_name, identifier_value in identifiers
    }

    return list(store.get_samples_by_any_id(**identifier_args))


def is_locked_attribute_on_sample(key: str, skip_attributes: List[str]) -> bool:
    """Returns true if the attribute is private or in the skip list."""
    return is_private_attribute(key) or key in skip_attributes


def is_private_attribute(key: str) -> bool:
    """Returns true if key has name indicating it is private."""
    return key.startswith("_")


def list_changeable_sample_attributes(
    sample: Optional[Sample] = None, skip_attributes: List[str] = []
) -> None:
    """List changeable attributes on sample and its current value."""
    LOG.info(f"Below is a set of changeable sample attributes, to combine with -kv flag:\n")

    sample_attributes: Iterable[str] = Sample.__dict__.keys()
    for attribute in sample_attributes:
        if is_locked_attribute_on_sample(attribute, skip_attributes):
            continue
        message: str = attribute
        if sample:
            message += f": {sample.__dict__.get(attribute)}"
        LOG.info(message)


@set_cmd.command()
@click.option("-s", "--sample_id", help="List all available modifiable keys for sample")
@click.pass_obj
def list_keys(
    context: CGConfig,
    sample_id: Optional[str],
):
    """List all available modifiable keys."""
    status_db: Store = context.status_db
    sample: Sample = status_db.get_sample_by_internal_id(internal_id=sample_id)
    list_changeable_sample_attributes(
        sample=sample, skip_attributes=NOT_CHANGEABLE_SAMPLE_ATTRIBUTES
    )


@set_cmd.command()
@click.argument("sample_id", required=True)
@click.option(
    OPTION_SHORT_KEY_VALUE,
    OPTION_LONG_KEY_VALUE,
    "kwargs",
    nargs=2,
    type=click.Tuple([str, str]),
    multiple=True,
    help=HELP_KEY_VALUE,
)
@click.option(OPTION_LONG_SKIP_LIMS, is_flag=True, help=HELP_SKIP_LIMS)
@click.option(OPTION_SHORT_YES, OPTION_LONG_YES, is_flag=True, help=HELP_YES)
@click.pass_obj
def sample(
    context: CGConfig,
    sample_id: Optional[str],
    kwargs: click.Tuple([str, str]),
    skip_lims: bool,
    yes: bool,
):
    """Set key values on a sample.

    \b
    To set apptag use: -kv application_version [APPTAG]
    To set customer use: -kv customer [CUSTOMER]
    To set priority use: -kv priority [priority as text or number]

    """
    status_db: Store = context.status_db
    sample: Sample = status_db.get_sample_by_internal_id(internal_id=sample_id)

    if sample is None:
        LOG.error(f"Can't find sample {sample_id}")
        raise click.Abort

    for key, value in kwargs:
        if is_locked_attribute_on_sample(key, NOT_CHANGEABLE_SAMPLE_ATTRIBUTES):
            LOG.warning(f"{key} is not a changeable attribute on sample")
            continue
        if not hasattr(sample, key):
            LOG.warning(f"{key} is not a property of sample")
            continue

        new_key: str = key
        if isinstance(getattr(sample, key), bool):
            new_value: bool = bool(value.lower() == "true")
        else:
            new_value: str = value

        if key in ["customer", "application_version", "priority"]:
            if key == "priority":
                if isinstance(value, str) and not value.isdigit():
                    new_key = "priority_human"
            elif key == "customer":
                new_value: Customer = status_db.get_customer_by_customer_id(customer_id=value)
            elif key == "application_version":
                new_value: ApplicationVersion = status_db.current_application_version(value)

            if not new_value:
                LOG.error(f"{key} {value} not found, aborting")
                raise click.Abort

        old_value = getattr(sample, new_key)

        LOG.info(f"Would change from {new_key}={old_value} to {new_key}={new_value} on {sample}")

        if not (yes or click.confirm(CONFIRM)):
            continue

        if key == "comment":
            _update_comment(new_value, sample)
        else:
            setattr(sample, new_key, new_value)
            _update_comment(_generate_comment(new_key, old_value, new_value), sample)

        status_db.commit()

    if not skip_lims:
        for key, value in kwargs:
            new_key = "application" if key == "application_version" else key
            new_value = sample.priority_human if key == "priority" else value
            LOG.info(f"Would set {new_key} to {new_value} for {sample.internal_id} in LIMS")

            if not (yes or click.confirm(CONFIRM)):
                raise click.Abort

            try:
                context.lims_api.update_sample(lims_id=sample_id, **{new_key: new_value})
                LOG.info(f"Set LIMS/{new_key} to {new_value}")
            except LimsDataError as error:
                LOG.error(f"Failed to set LIMS/{new_key} to {new_value}, {error}")


def _generate_comment(what, old_value, new_value):
    """Generate a comment that can be used in the comment field to describe updated value."""
    return f"\n{what} changed from {str(old_value)} to {str(new_value)}."


def _update_comment(comment, obj):
    """Appends the comment on obj including a timestamp"""
    if comment:
        timestamp = str(datetime.datetime.now())[:-10]
        if obj.comment is None:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}"
        else:
            obj.comment = f"{timestamp}-{getpass.getuser()}: {comment}" + "\n" + obj.comment


@set_cmd.command()
@click.option("-s", "--status", type=click.Choice(FLOWCELL_STATUS))
@click.argument("flowcell_name")
@click.pass_obj
def flowcell(context: CGConfig, flowcell_name: str, status: Optional[str]):
    """Update information about a flow cell."""
    status_db: Store = context.status_db
    flowcell_obj: Flowcell = status_db.get_flow_cell(flowcell_name)

    if flowcell_obj is None:
        LOG.warning(f"flow cell not found: {flowcell_name}")
        raise click.Abort
    prev_status: str = flowcell_obj.status
    flowcell_obj.status = status

    status_db.commit()
    LOG.info(f"{flowcell_name} set: {prev_status} -> {status}")


set_cmd.add_command(family)
set_cmd.add_command(families)



from typing import List, Callable
from enum import Enum
from sqlalchemy.orm import Query

from cg.constants import REPORT_SUPPORTED_PIPELINES
from cg.constants.constants import VALID_DATA_IN_PRODUCTION
from cg.store.models import Analysis
from cgmodels.cg.constants import Pipeline


def get_valid_analyses_in_production(analyses: Query, **kwargs) -> Query:
    """Return analyses with a valid data in production."""
    return analyses.filter(VALID_DATA_IN_PRODUCTION < Analysis.completed_at)


def get_analyses_with_pipeline(analyses: Query, pipeline: Pipeline = None, **kwargs) -> Query:
    """Return analyses with supplied pipeline."""
    return analyses.filter(Analysis.pipeline == str(pipeline)) if pipeline else analyses


def get_completed_analyses(analyses: Query, **kwargs) -> Query:
    """Return analyses that have been completed."""
    return analyses.filter(Analysis.completed_at.isnot(None))


def get_not_completed_analyses(analyses: Query, **kwargs) -> Query:
    """Return not completed analyses."""
    return analyses.filter(Analysis.completed_at.is_(None))


def get_filter_uploaded_analyses(analyses: Query, **kwargs) -> Query:
    """Return analyses that have been already uploaded."""
    return analyses.filter(Analysis.uploaded_at.isnot(None))


def get_not_uploaded_analyses(analyses: Query, **kwargs) -> Query:
    """Return analyses that have not been uploaded."""
    return analyses.filter(Analysis.uploaded_at.is_(None))


def get_analyses_with_delivery_report(analyses: Query, **kwargs) -> Query:
    """Return analyses that have a delivery report generated."""
    return analyses.filter(Analysis.delivery_report_created_at.isnot(None))


def get_analyses_without_delivery_report(analyses: Query, **kwargs) -> Query:
    """Return analyses that do not have a delivery report generated."""
    return analyses.filter(Analysis.delivery_report_created_at.is_(None))


def get_report_analyses_by_pipeline(analyses: Query, pipeline: Pipeline = None, **kwargs) -> Query:
    """Return the delivery report related analyses associated to the provided or supported pipelines."""
    return (
        analyses.filter(Analysis.pipeline == str(pipeline))
        if pipeline
        else analyses.filter(Analysis.pipeline.in_(REPORT_SUPPORTED_PIPELINES))
    )


def order_analyses_by_completed_at(analyses: Query, **kwargs) -> Query:
    """Return a query of ordered analyses (from old to new) by the completed_at field."""
    return analyses.order_by(Analysis.completed_at.asc())


def order_analyses_by_uploaded_at(analyses: Query, **kwargs) -> Query:
    """Return a query of ordered analyses (from old to new) by the uploaded_at field."""
    return analyses.order_by(Analysis.uploaded_at.asc())


def apply_analysis_filter(
    filter_functions: List[Callable], analyses: Query, pipeline: Pipeline = None
) -> Query:
    """Apply filtering functions to the analyses queries and return filtered results."""

    for function in filter_functions:
        analyses: Query = function(
            analyses=analyses,
            pipeline=pipeline,
        )
    return analyses


class AnalysisFilter(Enum):
    """Define Analysis filter functions."""

    FILTER_VALID_IN_PRODUCTION: Callable = get_valid_analyses_in_production
    FILTER_WITH_PIPELINE: Callable = get_analyses_with_pipeline
    FILTER_COMPLETED: Callable = get_completed_analyses
    FILTER_NOT_COMPLETED: Callable = get_not_completed_analyses
    FILTER_UPLOADED: Callable = get_filter_uploaded_analyses
    FILTER_NOT_UPLOADED: Callable = get_not_uploaded_analyses
    FILTER_WITH_DELIVERY_REPORT: Callable = get_analyses_with_delivery_report
    FILTER_WITHOUT_DELIVERY_REPORT: Callable = get_analyses_without_delivery_report
    FILTER_REPORT_BY_PIPELINE: Callable = get_report_analyses_by_pipeline
    ORDER_BY_COMPLETED_AT: Callable = order_analyses_by_completed_at
